mod execute;
#[cfg(windows)]
mod windows;

use std::fmt;

#[cfg(windows)]
use windows::service_main;
#[cfg(windows)]
use windows_service::define_windows_service;

use error_stack::{Context, Report};

use crate::service::execute::execute;

// Generate the windows service boilerplate.
// The boilerplate contains the low-level service entry function (ffi_service_main) that parses
// incoming service arguments into Vec<OsString> and passes them to user defined service
// entry (my_service_main).
#[cfg(windows)]
define_windows_service!(ffi_service_main, service_main);

#[derive(Debug)]
pub enum ServiceError {
    Starting,
    Installing,
    Uninstalling,
}
impl fmt::Display for ServiceError {
    fn fmt(&self, fmt: &mut fmt::Formatter<'_>) -> fmt::Result {
        fmt.write_str(&match self {
            Self::Starting => format!("Could not start {} service", env!("CARGO_PKG_NAME")),
            Self::Installing => format!(
                "Error on installing {} as a windows service",
                env!("CARGO_PKG_NAME")
            ),
            Self::Uninstalling => format!(
                "Error on installing {} as a windows service",
                env!("CARGO_PKG_NAME")
            ),
        })
    }
}
impl Context for ServiceError {}

#[cfg(not(windows))]
pub async fn run_service() -> Result<(), Report<ServiceStartingError>> {
    // Create a channel to be able to poll a stop event from the service worker loop.

    use std::sync::mpsc;
    let (_shutdown_tx, shutdown_rx) = mpsc::channel();

    // run service - blocking thread
    execute(shutdown_rx).await
}

#[cfg(windows)]
pub async fn run_service() -> Result<(), Report<ServiceError>> {
    // Get command line parameters

    use error_stack::{IntoReport, ResultExt};
    use windows_service::service_dispatcher;
    let args: Vec<String> = std::env::args().collect();

    let usage_output = format!(
        "
DaRL - Data and Runtime Logger\n
Usage: {}.exe service [OPTIONS]\n
OPTIONS:
install   \tCreate windows service
uninstall \tDelete windows service
",
        env!("CARGO_PKG_NAME")
    );
    match args.get(1).map(|s| s.as_str()) {
        Some("service") => match args.get(2).map(|s| s.as_str()) {
            Some("install") => match windows::install() {
                Ok(_) => println!("Successfully installed as service"),
                Err(e) => println!("Error installing service: {e:?}"),
            },

            Some("uninstall") => match windows::uninstall() {
                Ok(_) => println!("Successfully uninstalled"),
                Err(e) => println!("Error uninstalling service: {e:?}"),
            },
            Some(_) => println!("{usage_output}"),
            None => println!("{usage_output}"),
        },
        Some("--serviced") => {
            // Register generated `ffi_service_main` with the system and start the service, blocking
            // this thread until the service is stopped.
            service_dispatcher::start(env!("CARGO_PKG_NAME"), ffi_service_main)
                .into_report()
                .change_context(ServiceError::Starting)?;
        }
        Some(_) => println!("{usage_output}"),
        None => {
            use tokio::sync::mpsc;
            let (_shutdown_tx, shutdown_rx) = mpsc::channel(1);
            execute(shutdown_rx).await?;
        }
    };

    Ok(())
}

#[cfg(test)]
mod print_tests {

    #[test]
    fn print() {
        let usage_output = format!(
            "
DaRL - Data and Runtime Logger\n
Usage: {}.exe service [OPTIONS]\n
OPTIONS:
  install   \tCreate windows service
  uninstall \tDelete windows service
",
            env!("CARGO_PKG_NAME")
        );

        println!("{usage_output}");
    }
}
